/* squidclamav.c
   
   See the README file for usage instructions, and INSTALL
   for help on installing
   
*/


/*
  SquidClamav - A Clamav Antivirus Redirector for Squid

  Copyright (C) 2005-2010 Gilles Darold

        This program is free software: you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation, either version 3 of the License, or
        any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program.  If not, see < http://www.gnu.org/licenses/ >.

*/

#define TRUE  1
#define FALSE 0
#define MAXREDIR 10
#define VERSION "5.7"
#define COPYRIGHT  "Copyright (C) 2005-2010 Gilles Darold"
#define AUTHOR "Gilles Darold < gilles AT darold DOT net >"

#include "log.h"
#include "util.h"
#include "squidclamav.h"

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <ctype.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sysexits.h>
#include <errno.h>
#include <curl/curl.h>
#include <signal.h>
#include <sys/time.h>
#include <setjmp.h>
#include <sys/wait.h>

static char * escape_quote (char *s);
void timeit (struct timeval start, char *level);
int dconnect (void);
void replace_chr(char string[], char *from, char *to);
void free_global ();		/* routine to free global pointer */
void freeBuff (struct IN_BUFF);
/* routine to cleanup on signal */
void sigterm ();
int purge_cache(char *url);
int connectINET(char *serverHost, uint16_t serverPort);
void usage (void);

int bridge_mode = 0;
int interactive = 0;
int debug = 0;
int statit = 0;
int timeout = 10;
int wsockd = -1;
char *alternate_config = NULL;
char *redirect_url = NULL;
char *squidguard = NULL;
char *log_file = NULL;
char *proxy = NULL;
char *clamd_local = NULL;
char *clamd_ip = NULL;
char *clamd_port = NULL;
char *clamd_curr_ip = NULL;
char *useragent = NULL;
FILE *sgfpw = NULL;
FILE *sgfpr = NULL;
int maxredir = MAXREDIR;
int trust_cache = 0;
char *squid_ip = NULL;
char *squid_port = NULL;
char *sbuff = NULL; /* Squid entry */
SCPattern *patterns = NULL;
int pattc = 0;
int current_pattern_size = 0;
double maxsize = 0;
double worksize = 0;
int logredir = 0;
int nohead = 1;

regex_t *regexv = NULL;	/* Array of compiled patterns */
CURL *eh = NULL;	/* libcurl handler */
struct curl_slist *slist = NULL; 
struct IN_BUFF in_buff;

int skip_scan = 0;
int timeout_jump = 2;

void
reaper()
{
  while (wait3((union wait *) 0, WNOHANG, NULL) > 0);
}

/* Get download as stream  */
size_t
write_data (void *ptr, size_t data_size, size_t nmemb, void *stream)
{
  ssize_t ret;
  size_t msize = data_size * nmemb;

  if (debug > 1)
    logit(log_file, "DEBUG Sending data to clamd\n");
  ret = write (wsockd, ptr, msize);
  if (ret <= 0)
    {
	    logit(log_file, "ERROR Can't write to clamd socket.\n");
	    close (wsockd);
    }
  else
    {
      if (debug > 1)
	logit(log_file, "DEBUG Write %d bytes on %d to socket\n", ret, msize);
    }

  return msize;
}

/* Get squid header as stream  */
size_t
read_data (void *ptr, size_t data_size, size_t nmemb, void *stream)
{
  size_t msize = data_size * nmemb;

  if (debug > 2)
    logit(log_file, "DEBUG Received HTTP-HEADER: %s", (char *)ptr);
  /* Detect if the file come from Squid cache and if we may trust it*/
  if ((trust_cache == 1) && (strncmp(ptr, "X-Cache: HIT", 12) == 0) ) {
    skip_scan = 1;
  }
  if (maxsize > 0) {
     if ((strncmp(ptr, "Content-Length: ", 16) == 0) ) {
	if (sscanf(ptr, "Content-Length: %lf", &worksize) > 0) {
		if (worksize > maxsize) {
		   if (debug > 2)
		       logit(log_file, "DEBUG enabling Squid 2.X maxsize workaround: %s", (char *)ptr);
		    return -1;
		}
	}
     }
  }

  return msize;
}

/* 
    Under normal conditions, SquidClamav will never abort (you don't
    want your Squid to suddenly die if you've put an invalid
    regex into your config file, or renamed the config file!  
    Under these abnormal conditions, squidclamav will run in bridge_mode 
    where stdin is echoed to stdout!

    If you've entered an invalid regex expression that regcomp()
    doesn't like, or perhaps an invalid network number, check it 
    out in the ERROR_LOG.
    
    It should never ever abort under normal circumstances - even
    if it can't allocate memory (which shouldn't happen, and Squid 
    would be running like a dog under low memory conditions anyway) 
    bridge_mode will be set and it will keep running :-)
    
*/

int
main (int argc, char **argv)
{

  struct timeval g_start; /* Start global time structure */
  struct timeval cs_start; /* Start chaining and scanning time structure */
  int buff_status = 0;
  char cbuff[SMALL_BUFF]; /* Clamd return */
  char *escaped = NULL;

  double usize = 0;
  char *content_type = NULL;
  char err[CURL_ERROR_SIZE];
  int found = 0;
  int virusfound = 0;

  /* Clamd socket variable */
  int nbread = 0;
  int sockd;
  int loopw = 60;
  uint16_t port;
  struct sockaddr_in server;
  struct sockaddr_in peer;
  size_t peer_size;
  char *pt = NULL;
  FILE *chain_pipe = NULL;
  char *chain_ret = NULL;
  int pipe1[2];
  int pipe2[2];
  pid_t pid;
  int usepipe = 0;
  struct timespec sleeptime;
  CURLcode res ;
  char *ret;
  ssize_t rret;
 
  /* Add support to command line options */
  int next_option = 0;
  const char* const short_options = "c:d:hv";
  const struct option long_options[] = {
    { "config",   1, NULL, 'c' },
    { "debug",    1, NULL, 'd' },
    { "help",     0, NULL, 'h' },
    { "version",  0, NULL, 'v' },
    { NULL,       0, NULL, 0   }
  };

  sleeptime.tv_sec = 0;
  sleeptime.tv_nsec = 1000000;

  timeout_jump = atoi(CLAMD_TIMEOUT);

  do {
    nanosleep(&sleeptime, NULL);		/* don't do DoS */
    next_option = getopt_long(argc, argv, short_options, long_options, NULL);
    switch (next_option)
    {
    case 'h':   /* -h or --help */
	/* print usage information. */
	usage();
	exit(0);

    case 'c':   /* -c or --config */
	/* This option takes an argument, the name of the output file.  */
	alternate_config = (char *)malloc(sizeof(char) * LOW_BUFF);
	if(alternate_config == NULL) {
		fprintf(stderr, "unable to allocate memory in main()\n");
		exit(EXIT_FAILURE);
	}
	xstrncpy(alternate_config, optarg, LOW_BUFF);
	break;

    case 'd':   /* -d or --debug */
	debug = atoi(optarg);
	break;

    case 'v':   /* -v or --version */
	fprintf(stderr, "SquidClamav v%s\n", VERSION); 
	exit(0);

    case '?':   /* The user specified an invalid option.  */
	usage();
	exit(0);

    case -1:    /* Done with options.  */
      break;

    default:    /* Something else: unexpected.  */
	usage();
	exit(0);
    }
  }
  while (next_option != -1);

  /* go into interactive mode if we're run as root */
  if ((int) getuid () == 0)
    {
      interactive = 1;
      fprintf (stderr, "SquidClamav v%s running in interactive mode\n", VERSION);
    }

  /* make stdout line buffered */
  if (setvbuf (stdout, (char *)NULL, _IOLBF, 0) != 0)
    {
      fprintf (stderr, "stdout not line buffered exiting\n");
      nanosleep(&sleeptime, NULL);		/* don't do DoS */
      exit (EXIT_FAILURE);
    }

  bridge_mode = 0;
  pattc = 0;
  current_pattern_size = 0;

  signal(SIGHUP, sigterm);	/* set function calls on signal trap */
  signal (SIGTERM, sigterm);
  signal (SIGUSR1, sigterm);
  signal(SIGCHLD, reaper);

 /*********************
   read config files
  ********************/
  load_patterns();

  if (bridge_mode == 1)
    logit(log_file, "PANIC Invalid condition - continuing in BRIDGE mode\n");

  logit(log_file, "LOG SquidClamav v%s (PID %d) started\n", VERSION, (int) getpid ());

  if (bridge_mode == 0) {

      /* curl init */
      curl_global_init (CURL_GLOBAL_ALL);
      /* get an easy handle */
      if ((eh = curl_easy_init ()) == NULL)
	{
	  logit(log_file, "PANIC Curl easy initialization failed\n");
	  curl_global_cleanup ();
	  bridge_mode = 1;
	  logit(log_file, "PANIC Invalid condition - continuing in BRIDGE mode\n");
	}
      else
	{
	  /* set the curl output header redirected to a function */
	  curl_easy_setopt (eh, CURLOPT_HEADERFUNCTION, &read_data);
	  /* set the curl output to temporary file */
	  curl_easy_setopt (eh, CURLOPT_WRITEFUNCTION, &write_data);
	  /* set the error buffer */
	  curl_easy_setopt (eh, CURLOPT_ERRORBUFFER, err);
	  /* do not install signal  handlers in thread context */
	  curl_easy_setopt (eh, CURLOPT_NOSIGNAL, 1);
	  /* Force HTTP 1.0 version */
	  curl_easy_setopt (eh, CURL_HTTP_VERSION_1_0, TRUE);
	  /* Do not fail on errors: use to prevent breaked download when
	   * size of the download is upper than clamd scan limit
	   * Drop error on debug mode  */
	  if (debug != 0) {
		  curl_easy_setopt (eh, CURLOPT_FAILONERROR, TRUE);
	  } else {
		  curl_easy_setopt (eh, CURLOPT_FAILONERROR, FALSE);
	  }
	  /* follow location (303 MOVED) */
	  curl_easy_setopt (eh, CURLOPT_FOLLOWLOCATION, TRUE);
	  /* only follow 3 redirects */
	  curl_easy_setopt (eh, CURLOPT_MAXREDIRS, maxredir);
	  /* simulates being another user agent */
	  if (useragent != NULL)
		curl_easy_setopt (eh, CURLOPT_USERAGENT, useragent);
	  /* overwrite the Pragma: no-cache HTTP header */
	  slist = curl_slist_append(slist, "pragma:");
	  curl_easy_setopt (eh, CURLOPT_HTTPHEADER, slist);
	  /* Suppress error: SSL certificate problem, verify that the CA cert is OK */
	  curl_easy_setopt (eh, CURLOPT_SSL_VERIFYHOST, 0);
	  curl_easy_setopt (eh, CURLOPT_SSL_VERIFYPEER, 0);
	}
   }
  /* create a squidguard child process and setup pipes */
  if ((bridge_mode == 0) && (squidguard != NULL)) {

	if ( pipe(pipe1) < 0  ||  pipe(pipe2) < 0 )
	{
		logit(log_file,"ERROR unable to open pipe, disabling call to %s.\n", squidguard);
		usepipe = 0;
	} else {
		if ( (pid = fork()) == -1) {
			logit(log_file,"ERROR unable to fork, disabling call to %s.\n", squidguard);
			usepipe = 0;
		} else {
			if(pid == 0) {
				close(pipe1[1]);
				dup2(pipe1[0],0);
				close(pipe2[0]);
				dup2(pipe2[1],1);
				setsid();
				/* Running chained program */
				execlp(squidguard,(char *)basename(squidguard),(char  *)0);
				exit(EXIT_SUCCESS);
				return(0);
			} else {
				close(pipe1[0]);
				sgfpw = fdopen(pipe1[1], "w");
				if(!sgfpw) {
					logit(log_file,"ERROR unable to fopen squidGuard childs stdin, disabling squidGuard.\n");
					usepipe = 0;
				} else {
				        /* make pipe line buffered */
					if (setvbuf (sgfpw, (char *)NULL, _IOLBF, 0)  != 0)
						logit(log_file,"WARNING unable to line buffering pipe.\n");
					sgfpr = fdopen(pipe2[0], "r");
					if(!sgfpr) {
						logit(log_file,"ERROR unable to fopen squidGuard childs stdout, disabling squidGuard.\n");
						usepipe = 0;
					} else {
						logit(log_file,"bidirectional pipe to %s childs ready...\n", squidguard);
						usepipe = 1;
					}
				}
			}
		}
	}
  }


  /* Loop on squid inout */
  clamd_curr_ip = (char *) malloc (sizeof (char) * LOW_CHAR);
  memset(clamd_curr_ip, 0, sizeof(clamd_curr_ip));
  sbuff = (char *)malloc(sizeof(char)*MAX_URL); /* Squid entry */
  while (fgets(sbuff, MAX_URL, stdin) != NULL) {

	/* chop sbuff */
	chomp(sbuff);
	worksize = 0;

	/* Start timing */
	if (statit == 1)
		gettimeofday(&g_start,NULL);

	/* Separate Squid input into 4 fields */
	buff_status = load_in_buff(sbuff);

	/* if four fields couldn't be separated, or the converted values
	   aren't appropriate, then just echo back the line from stdin */
	if (buff_status == 1) {
	  logit(log_file, "DEBUG Invalid input buffer, aborting: %s\n", sbuff);
	  puts("");
	  fflush(stdout);
	  continue;
	}

      if ((bridge_mode == 0) && (buff_status == 0) ) {
        /* if a TRUSTUSER match => no squidguard and no virus scan */
	found = simple_pattern_compare (in_buff.ident, TRUSTUSER);
	if (found == 1) {
	  if (debug > 1)
	    logit(log_file, "DEBUG No squidguard and no antivir check (TRUSTUSER match) for user: %s\n", in_buff.ident);
	  if (statit == 1)
		timeit(g_start, "Total");
	  puts("");
	  fflush(stdout);
	  continue;
	}

	/* if a TRUSTCLIENT match => no squidguard and no virus scan */
	found = client_pattern_compare(in_buff.ipaddress, in_buff.fqdn);
	if (found > 0) {
	  if (debug > 1)
	    logit(log_file, "DEBUG No squidguard and no antivir check (TRUSTCLIENT match) for address: %s/%s\n",in_buff.ipaddress, in_buff.fqdn);
	  if (statit == 1)
		timeit(g_start, "Total");
	  puts("");
	  fflush(stdout);
	  continue;
	}

        /* if a WHITELIST match => no squidguard and no virus scan */
	found = simple_pattern_compare (in_buff.url, WHITELIST);
	if (found == 1) {
	  if (debug > 1)
	    logit(log_file, "DEBUG No squidguard and no antivir check (WHITELIST match) for url: %s\n", in_buff.url);
	    if (statit == 1)
		timeit(g_start, "Total");
	  puts("");
	  fflush(stdout);
	  continue;
	}
      }

      /* chaining with SquidGuard - before bridge mode or not*/
      if ((bridge_mode == 0) && (squidguard != NULL)) {
		if (usepipe == 1) {
			if (debug > 0)
			    logit(log_file, "DEBUG Sending request to chained program: %s\n", squidguard);
			fprintf(sgfpw,"%s\n",sbuff);
			fflush(sgfpw);
			xfree(escaped);
			escaped = NULL;
			/* the chained redirector must return empty line if ok or the redirection url */
			chain_ret = (char *)malloc(sizeof(char)*MAX_URL);
			ret = fgets(chain_ret,MAX_URL,sgfpr);
			if ((ret != NULL) && (strlen(chain_ret) > 1)) {
				if (logredir)
				    logit(log_file, "Chained program redirection (%s/%s): %s", in_buff.ident, in_buff.ipaddress, chain_ret);
				if (statit == 1)
					timeit(g_start, "Chained program call");
				/* return the redirection url to squid */
				fprintf (stdout, "%s", chain_ret);
				fflush(stdout);
				/* do not go further */
				xfree(chain_ret);
				chain_ret = NULL;
				continue;
			}
			xfree(chain_ret);
			chain_ret = NULL;
		}
		if (statit == 1)
			timeit(g_start, "Chained program call");
       }

       /* if configs are completely invalid or some other
         exception occurs where we want the redirector to
         continue operation (so that Squid still works!),
         we simply echo stdin to stdout - i.e. "bridge mode" :-) */
       if (bridge_mode == 1) {
	  puts("");
	  fflush(stdout);
	  continue;
       }

	/* if four fields couldn't be separated, or the converted values
	   aren't appropriate, then just echo back the line from stdin */
	if (buff_status == -1) {
	  if (debug > 2)
	     logit(log_file, "DEBUG method is not GET skipping virus scan.\n");
	  puts("");
	  fflush(stdout);
	  continue;
	}

      /* Return 1 mean url ABORT match => No antivir check */
      found = simple_pattern_compare(in_buff.url, ABORT);
      if (found == 1) {
	  if (debug > 1)
	    logit(log_file, "DEBUG No antivir check (ABORT match) for url: %s\n", in_buff.url);
	  if (statit == 1)
		timeit(g_start, "Total");

	  /* no replacement for the URL was found */
	  puts("");
	  fflush(stdout);
	  continue;

      } else {

	  skip_scan = 0;

	  /* set proxy */
	  if (strcmp(proxy,"none") != 0) {
		  if (debug > 0)
		    logit(log_file, "DEBUG Curl will use proxy: %s\n", proxy);
		curl_easy_setopt (eh, CURLOPT_PROXY, proxy);
	  }
	  /* set the url */
	  curl_easy_setopt (eh, CURLOPT_URL, in_buff.url);
	  /* set the libcurl transfer timeout to max 2 second for header */
	  curl_easy_setopt (eh, CURLOPT_TIMEOUT, atoi(CURL_TIMEOUT));
	  /* disable ssl verify */
	  curl_easy_setopt (eh, CURLOPT_SSL_VERIFYPEER, FALSE);
	  curl_easy_setopt (eh, CURLOPT_SSL_VERIFYHOST, FALSE);
	  /* only get the header to check size and content type */
	  curl_easy_setopt (eh, CURLOPT_NOBODY, TRUE);

	  /* download the header only if abortcontent is set in config file */
	if (nohead == 0) {
         if (debug > 1)
		logit(log_file, "DEBUG looking for Content-Type of url %s\n", in_buff.url);
	  /* if (curl_easy_perform (eh) != 0) */
	  res = curl_easy_perform (eh);
	  if (res != 0) {
		if (debug != 0)
		   logit(log_file, "DEBUG bad header (%d: %s)\n", res, err);
	        if (res == 7) {
		      /* Can't connect to host */
		      if (debug != 0)
		        logit(log_file, "DEBUG can't connect to host, aborting scan.\n");
		      continue;
		} else {
		   if (strcmp(err, "The requested URL returned error: 404") != 0) {
		      /* Bad header, just warn */
		      if (debug != 0)
		        logit(log_file, "DEBUG bad header (%d: %s)\n", res, err);
		   }
		}
	  }
	  if (skip_scan == 1) {
	      if (debug > 0)
		   logit(log_file, "DEBUG HIT Cache found, trust cache enabled, skipping...\n");
	      if (statit == 1)
		timeit(g_start, "Total");
	      puts("");
	      fflush(stdout);
	      continue;
	  }

	  /* get the Content Length from the header */
	  if (curl_easy_getinfo (eh, CURLINFO_CONTENT_LENGTH_DOWNLOAD, &usize) != CURLE_OK) {
		      /* not ok --> squid should handle this */
		      logit(log_file, "ERROR No content length from url %s\n", in_buff.url);
		      if (statit == 1)
				timeit(g_start, "Total");
		      puts("");
		      fflush(stdout);
		      continue;
	  }
	  if (debug > 1)
	    logit(log_file, "DEBUG File size is %'.2f\n", usize);

	  /* Do not virus scan if maxsize is reach */
	  if ( (maxsize > 0) && (usize >= maxsize) ) {
	     if (debug > 0)
		      logit(log_file, "DEBUG No antivir check (Content length is upper than maxsize): %'.2f > %'.2f\n", usize, maxsize);
	     if (statit == 1)
		timeit(g_start, "Total");
	     puts("");
	     fflush(stdout);
	     continue;
	  }

	  /* get the CONTENT_TYPE out of the header */
	  if (curl_easy_getinfo (eh, CURLINFO_CONTENT_TYPE, &content_type) != CURLE_OK) {
		  /* not ok --> squid should handle this */
		  logit(log_file, "ERROR No content type from url %s\n",
		       in_buff.url);
		  if (statit == 1)
			timeit(g_start, "Total");
		  puts("");
		  fflush(stdout);
		  continue;
	  }
	  /* fix headers coming with \r terminaison */
	  if (content_type != NULL)
		chomp(content_type);
	  if ((content_type != NULL) && (simple_pattern_compare(content_type, ABORTCONTENT) == 1)) {
		if (debug != 0)
		    logit(log_file, "DEBUG No antivir check (ABORTCONTENT match) for content-type: %s\n", content_type);
		if (statit == 1)
			timeit(g_start, "Total");
		puts("");
		fflush(stdout);
		continue;
	   }
	   if (debug > 0)
		logit(log_file, "DEBUG Content-Type: %s will be scanned\n", content_type);
	}
	  /* Start timing */
	  if (statit == 1)
		  gettimeofday(&cs_start,NULL);
	  /***** Open a socket to Clamd daemon *****/
	  if ((sockd = dconnect ()) < 0)
	    {
	      logit(log_file, "ERROR Can't connect to Clamd daemon, fallback to Squid.\n");
	      puts("");
	      fflush(stdout);
	      continue;
	    }
	  if (debug != 0)
	    logit(log_file, "DEBUG Sending STREAM to clamd.\n");

	  if (write (sockd, "STREAM", 6) <= 0)
	    {
	      logit(log_file, "ERROR Can't write to Clamd socket.\n");
	      puts("");
	      fflush(stdout);
	      continue;
	    }

	  while (loopw > 0) {
	      memset(cbuff, 0, sizeof(cbuff));
	      rret = read (sockd, cbuff, LOW_BUFF);
	      if ((rret > -1) && (pt = strstr (cbuff, "PORT"))) {
		  pt += 5;
		  sscanf (pt, "%d", (int *)&port);
		  break;
		}
	      loopw--;
	    }
	  if (loopw == 0)
	    {
	      logit(log_file, "ERROR Clamd daemon not ready for stream scanning.\n");
	      puts("");
	      fflush(stdout);
	      continue;
	    }
	  if (debug > 1)
	    logit(log_file, "DEBUG Received port %d from clamd.\n", port);

	  /* connect to clamd given port */
	  if ((wsockd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
	    {
	      logit(log_file, "ERROR Can't create the Clamd socket.\n");
	      puts("");
	      fflush(stdout);
	      continue;
	    }

	  server.sin_family = AF_INET;
	  server.sin_port = htons (port);

	  peer_size = sizeof (peer);

	  if (getpeername (sockd, (struct sockaddr *) &peer, (socklen_t *) &peer_size) < 0)
	    {
	      logit(log_file, "ERROR Can't get socket peer name.\n");
	      puts("");
	      fflush(stdout);
	      continue;
	    }
	  switch (peer.sin_family)
	    {
	    case AF_UNIX:
	      server.sin_addr.s_addr = inet_addr ("127.0.0.1");
	      break;
	    case AF_INET:
	      server.sin_addr.s_addr = peer.sin_addr.s_addr;
	      break;
	    default:
	      logit(log_file, "ERROR Unexpected socket type: %d.\n",
		   peer.sin_family);
	      continue;
	    }

	  if (debug > 1)
	    logit(log_file, "DEBUG Trying to connect to clamd [port: %d].\n", port);

	  if (connect (wsockd, (struct sockaddr *) &server,
	       sizeof (struct sockaddr_in)) < 0)
	    {
	      close (wsockd);
	      logit(log_file, "ERROR Can't connect to clamd [port: %d].\n", port);
	      puts("");
	      fflush(stdout);
	      continue;
	    }

	  /***** now get the body content *****/
	  /* download body */
	  curl_easy_setopt (eh, CURLOPT_NOBODY, FALSE);
	  curl_easy_setopt (eh, CURLOPT_HTTPGET, TRUE);
	  /* set the libcurl transfer timeout */
	  curl_easy_setopt (eh, CURLOPT_TIMEOUT, timeout);
	  res = curl_easy_perform (eh);
	  if (res != 0) {
	      if ((worksize > 0) && (worksize < maxsize)) {
		      logit(log_file, "ERROR fail downloading url %s\n", in_buff.url);
		      logit(log_file, "ERROR CURLOPT_ERRORBUFFER: %s\n", err);
	      }
	      close (wsockd);
	      close (sockd);
	      if (debug > 1)
	         logit(log_file, "DEBUG Connection to clamd on port: %d closed.\n", port);
	      puts("");
	      fflush(stdout);
	      if (statit == 1)
		 timeit(g_start, "Total");
	      continue;
	  }
	  close (wsockd);
	  if (debug > 1)
	    logit(log_file, "DEBUG Connection to clamd on port: %d closed.\n", port);
	  if (debug > 1)
	    logit(log_file, "DEBUG Reading clamd scan result.\n");
	  memset(cbuff, 0, sizeof(cbuff));
	  virusfound = 0;
	  while ((nbread = read(sockd, cbuff, SMALL_BUFF)) > 0) {
	      if (debug > 1)
		logit(log_file, "DEBUG received from Clamd: %s", cbuff);
	      if (strstr (cbuff, "FOUND\n"))
		{
		  char *urlredir = (char *)malloc(sizeof(char)*MAX_URL);
		  char *badurl = (char *)malloc(sizeof(char)*NORMAL_BUFF);
		  xstrncpy(badurl, in_buff.url, NORMAL_BUFF);
		  snprintf(urlredir, MAX_URL, "%s?url=%s&source=%s&user=%s&virus=%s",
			redirect_url,badurl,in_buff.src_address,in_buff.ident,cbuff);
		  xfree(badurl);
		  replace_chr(urlredir, " ", "+");
		  /* chop urlredir */
		  chomp(urlredir);
		  logit(log_file, "LOG Redirecting URL to: %s\n", urlredir);
		  virusfound = 1;
		  if (trust_cache == 1) {
			if (debug != 0)
			   logit(log_file, "DEBUG Virus found, trying to purge Squid Cache.\n");
			if (purge_cache(in_buff.url) == 0) {
			   trust_cache = 0;
			} else {
			   logit(log_file, "Squid Cache purged of url %s.\n", in_buff.url);
			}
		  }
		  fprintf (stdout, "%s %s %s %s\n", urlredir,
		       in_buff.src_address, in_buff.ident, in_buff.method);
		  fflush(stdout);
		  xfree(urlredir);
		  if (debug != 0)
			logit(log_file, "DEBUG Virus found, redirection sent to Squid.\n");
		}
	        memset(cbuff, 0, sizeof(cbuff));
	  }
	  close (sockd);
	  if (debug > 1)
	    logit(log_file, "DEBUG End reading clamd scan result.\n");
	  if (statit == 1)
		timeit(cs_start, "Virus Scanning");

	  if (virusfound == 0) {
	      if (debug != 0)
		logit(log_file, "DEBUG No virus detected.\n");
	      puts("");
	      fflush(stdout);
	  }
	}
	if (statit == 1)
		timeit(g_start, "Total");
    }

  /* cleanup */
  xfree(escaped);
  free_global();
  while(wait(NULL)>0);

  return(EXIT_SUCCESS);
}

void
sigterm ()
{
  signal(SIGTERM, sigterm);	/* reset signal */
  logit(log_file, "LOG Signal received, cleaning all before quit...\n");
  /* cleanup */
  xfree(alternate_config);
  free_global();
  while(wait(NULL)>0);
  exit(EXIT_SUCCESS);
}

void
free_global ()
{
  xfree(alternate_config);
  xfree(clamd_ip);
  xfree(clamd_port);
  xfree(clamd_curr_ip);
  xfree(clamd_local);
  xfree(squid_ip);
  xfree(squid_port);
  xfree(proxy);
  xfree(redirect_url);
  xfree(squidguard);
  xfree(log_file);
  xfree(useragent);
  xfree(sbuff);
  if (patterns != NULL) {
	while (pattc > 0) {
		pattc--;
		regfree(&patterns[pattc].regexv);
		xfree(patterns[pattc].pattern);
	}
	free(patterns);
  }
  curl_slist_free_all(slist);
  curl_easy_cleanup(eh);
  curl_global_cleanup ();
  if (sgfpw) fclose(sgfpw);
  if (sgfpr) fclose(sgfpr);
}

int
dconnect ()
{
  struct sockaddr_un userver;
  int asockd;


  memset ((char *) &userver, 0, sizeof (userver));

  if (clamd_local != NULL)
    {
      userver.sun_family = AF_UNIX;
      xstrncpy (userver.sun_path, clamd_local, sizeof(userver.sun_path));
      if ((asockd = socket (AF_UNIX, SOCK_STREAM, 0)) < 0)
	{
	  logit(log_file, "ERROR Can't bind local socket on %s.\n", clamd_local);
	  return -1;
	}
      if (connect (asockd, (struct sockaddr *) &userver, sizeof (struct sockaddr_un)) < 0) {
	  close (asockd);
	  logit(log_file, "ERROR Can't connect to clamd on local socket %s.\n", clamd_local);
	  return -1;
      }
      return asockd;

    } else {	

	if (clamd_curr_ip[0] != 0) {
		asockd = connectINET(clamd_curr_ip, atoi(clamd_port));
		if ( asockd != -1 ) {
		    if (debug > 1)
			logit(log_file, "Connected to Clamd (%s:%s)\n", clamd_curr_ip,clamd_port);
		    return asockd;
		}
	}
	char *ptr;
	char *s = (char *) malloc (sizeof (char) * SMALL_CHAR);
	xstrncpy(s, clamd_ip, SMALL_CHAR);
	ptr = strtok(s, ",");
	while (ptr != NULL) {
		asockd = connectINET(ptr, atoi(clamd_port));
		if ( asockd != -1 ) {
		    if (debug > 1)
			logit(log_file, "Connected to Clamd (%s:%s)\n", ptr,clamd_port);
		    /* Store last working clamd */
		    xstrncpy(clamd_curr_ip, ptr, LOW_CHAR);
		    xfree(s);
		    break;
		}
		ptr = strtok(NULL, ",");
	}
	return asockd;
	xfree(s);
    }
    return 0;
}

void
replace_chr (char string[], char *from, char *to)
{
  char *start, *p1, *p2;
  for (start = string; *start != '\0'; start++)
    {
      p1 = from;
      p2 = start;
      while (*p1 != '\0')
	{
	  if (*p1 != *p2)
	    break;
	  p1++;
	  p2++;
	}
      if (*p1 == '\0')
	{
	  for (p1 = to; *p1 != '\0'; p1++)
	    *start++ = *p1;
	}
    }
}

void
timeit (struct timeval start, char *level)
{
	struct timeval stop;

	/* Stop timing and compute elapsed time */
	gettimeofday(&stop,NULL);
	/* show stat */
	logit(log_file, "STAT %s process time %7.3f second(s)\n", level, (stop.tv_sec + (stop.tv_usec * 0.000001)) - (start.tv_sec + (start.tv_usec * 0.000001)));
}

/* 
   function to compute new string length,
   allocate memory and finally copy string 
   - Returned value must be freed -
 */
static char *
escape_quote (char *s)
{
	char *newstring;
	char *idx1, *idx2;
	int nbquotes = 0;
	size_t len = 0;

	if (!s)
		return;

	len = xstrnlen (s, MAX_URL);

	/* First, count quotes in string */
	idx1 = s;
	while (*idx1) {
		if ((*(idx1) == '"') || (*(idx1) == '\\') || (*(idx1) == '$'))
			nbquotes++;
		idx1++;
	}
	/* If there is no quote in the string, return it */
	if (nbquotes == 0)
		return (strdup (s));

	/* Or else, allocate memory for the new string */
	newstring = (char *) malloc (sizeof (char) * (len + nbquotes + 1));
	if (newstring == NULL) {
		logit(log_file, "PANIC Unable to allocate memory in escape_quote()\n");
		return;
	}
	/* And prepend a backslash before any quote found in string */
	idx1 = s;
	idx2 = newstring;
	while (*idx1) {
		if ((*(idx1) == '"') || (*(idx1) == '\\') || (*(idx1) == '$')) {
			*(idx2++) = '\\';
		}
		*(idx2++) = *idx1++;
	}
	/* Add null char */
	*idx2 = '\0';

	return newstring;
}

int
purge_cache(char *url)
{
	static const char* purge_cmd = "PURGE %s HTTP/1.0\r\nAccept: */*\r\n\r\n";
	char* buffer = NULL;
	unsigned long buff_size = strlen(url) + strlen(purge_cmd) + 4;
	int sockfd;
	size_t size = 0;
	long int result = 0;

	buffer = (char *)malloc(buff_size);
	snprintf( buffer, buff_size, purge_cmd, url );
	sockfd = connectINET(squid_ip, atoi(squid_port));
	if ( sockfd == -1 ) {
		logit(log_file,"ERROR: unable to connect to Squid: %s\n", strerror(errno) );
		xfree(buffer);
		return 0;
	}

	size = strlen(buffer);
	if ( write( sockfd, buffer, size ) != size ) {
		logit(log_file,"ERROR: unable to talk to squid: %s\n", strerror(errno) );
		close(sockfd);
		xfree(buffer);
		return 0;
	}
	memset( buffer+8, 0, 4 );
	if ( read( sockfd, buffer, buff_size ) < 1 ) {
		logit(log_file,"ERROR: unable to read squid answer: %s\n", strerror(errno) );
		close(sockfd);
		xfree(buffer);
		return 0;
	}
	close(sockfd);
	result = strtol(buffer+8,0,10);
	if (debug > 0)
	   logit(log_file,"DEBUG Purge squid answer: %ld\n", result);
	if (result == 403) {
	      logit(log_file,"ERROR Purge method not allowed. Please edit your squid ACL.\n");
	      logit(log_file,"ERROR Fallback to untrusted cache.\n");
	      trust_cache = 0;
	}
	xfree(buffer);

	return 1;
}

void
connect_timeout() {
   // doesn't actually need to do anything
}

int
connectINET(char *serverHost, uint16_t serverPort)
{
	struct sockaddr_in server;
	struct hostent *he;
	int asockd;
	struct sigaction action;
	action.sa_handler = connect_timeout;

	memset ((char *) &server, 0, sizeof (server));

	server.sin_addr.s_addr = inet_addr(serverHost);
	if ((asockd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
	{
	  logit(log_file, "ERROR Can't create a socket.\n");
	  return -1;
	}

	server.sin_family = AF_INET;
	server.sin_port = htons(serverPort);

	if ((he = gethostbyname(serverHost)) == 0)
	{
	  close (asockd);
	  logit(log_file, "ERROR Can't lookup hostname of %s\n", serverHost);
	  return -1;
	}
	server.sin_addr = *(struct in_addr *) he->h_addr_list[0];

	sigaction(SIGALRM, &action, NULL);
	alarm(timeout_jump);
	if (connect (asockd, (struct sockaddr *) &server, sizeof (struct sockaddr_in)) < 0) {
	  close (asockd);
	  logit(log_file, "ERROR Can't connect on %s:%d.\n", serverHost,serverPort);
	  return -1;
	}
	int err = errno;
	alarm(0);
	if (err == EINTR) {
		close(asockd);
	  	logit(log_file, "ERROR connectINET: Timeout connecting to clamd on %s:%d.\n", serverHost,serverPort);
	}


	return asockd;
}

/* Dumps SquidClamav usage information to stderr */
void
usage (void)
{
  fprintf(stderr, "Usage:  squidclamav [-v] [-h] [-c config file]\n");
  fprintf (stderr,
	"  -c  --config filename  Where alternate config file can be found.\n"
	"  -d  --debug  level     Set debug level. Possible values are 1 to 3.\n"
	"  -h  --help             Display this usage information.\n"
	"  -v  --version          Print version and exit.\n\n"
	"%s\n"
	"%s\n\n", COPYRIGHT, AUTHOR);
}


